<?php

namespace Drupal\funnelback;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\RfcLogLevel;
use Drupal\Core\Messenger\MessengerInterface;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\RequestException;

/**
 * Service class for funnelback client.
 */
class FunnelbackClient {

  /**
   * Debug level: none.
   *
   * @var string
   */
  protected $debugNone = 'none';

  /**
   * Debug level: basic logging.
   *
   * @var string
   */
  protected $debugLog = 'log';

  /**
   * Debug level: comprehensive logging.
   *
   * @var string
   */
  protected $debugVerbose = 'verbose';

  /**
   * The query string service.
   *
   * @var \Drupal\funnelback\FunnelbackQueryString
   */
  protected $funnelbackQueryString;

  /**
   * The config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The logger factory service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $logger;

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * FunnelbackClient constructor.
   *
   * @param \Drupal\funnelback\FunnelbackQueryString $funnelbackQueryString
   *   The query string service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger
   *   The logger factory service.
   * @param \GuzzleHttp\Client $httpClient
   *   The HTTP client.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   */
  public function __construct(FunnelbackQueryString $funnelbackQueryString, ConfigFactoryInterface $configFactory, LoggerChannelFactoryInterface $logger, Client $httpClient, MessengerInterface $messenger) {
    $this->funnelbackQueryString = $funnelbackQueryString;
    $this->configFactory = $configFactory;
    $this->logger = $logger;
    $this->httpClient = $httpClient;
    $this->messenger = $messenger;
  }

  /**
   * Make a request.
   *
   * @param string $baseUrl
   *   The base URL for the request.
   * @param string $apiPath
   *   The api path from the base URL.
   * @param array $requestParams
   *   The request parameters.
   *
   * @return object
   *   The response object.
   */
  public function request($baseUrl, $apiPath, array $requestParams) {

    $response = NULL;

    if ($baseUrl !== '/') {
      // Build the search URL with query params.
      $url = $baseUrl . $apiPath . '?' . http_build_query($requestParams);

      $url = $this->funnelbackQueryString->funnelbackQueryNormaliser($url);

      // Make the request.
      try {
        $response = $this->httpClient->request('get', $url);

        $this->debug('Requesting url: %url. Response %response. Template %template', [
          '%url' => $url,
          '%response' => $response->getStatusCode(),
          '%template' => ($apiPath == 's/search.json') ? t('Default') : t('Custom'),
        ]);
      }
      catch (RequestException $e) {
        $this->debug('Fail to connect to Funnelback server.');
      }
    }

    return $response;
  }

  /**
   * Helper to log debug messages.
   *
   * @param string $message
   *   A message, suitable for watchdog().
   * @param array $args
   *   (optional) An array of arguments, as per watchdog().
   * @param int $logLevel
   *   (optional) The watchdog() log level. Defaults to WATCHDOG_DEBUG.
   */
  public function debug($message, array $args = [], $logLevel = 7) {

    $config = $this->configFactory->get('funnelback.settings');
    $debug = $config->get('general_settings.debug_mode');
    $string = new FormattableMarkup($message, $args);
    if ($debug == $this->debugLog) {

      $this->logger->get('funnelback')->notice($string);
    }
    elseif ($debug == $this->debugVerbose) {
      if ($logLevel >= RfcLogLevel::ERROR) {
        $messageLevel = 'error';
      }
      else {
        $messageLevel = 'status';
      }
      $this->messenger->addMessage($string, $messageLevel);
    }
  }

}
